"""
The solution for Part 5.
"""

from collections import Counter

from part2 import friend_besties
from part3 import friend_second_besties

def predict_feature(feature: str, people: list[str],
    feat_dict: dict[str, dict[str, str]]) -> list[str] | None:
    """Predicts the value of a feature given a list of people to search through.

    ### Parameters:
        feature (str): The feature of interest to predict.
        people (list): The list of people to search through.
        feat_dict (dict[str, dict[str, str]]): The list of features for the
        people.

    ### Returns:
        feature (list or None): Returns a sorted list of predictions of features
        if people had that attribute. Otherwise returns None.
    """

    # Create a frequency dictionary of the values that the besties have for
    # this feature
    values = []
    for person in people:
        value = feat_dict[person][feature] if feature in feat_dict[person] else None
        if value is not None:
            values.append(value)
    freqs = Counter(values).most_common()

    # None of the besties had this attribute, so continue
    if len(freqs) == 0:
        return None

    # Only one bestie had this attribute, so it becomes the predicted value
    if len(freqs) == 1:
        [(value, _)] = freqs
        return value

    # Get the list of most common elements including ties
    return list(sorted([i for (i, v) in freqs if v == freqs[0][1]]))

def friendly_prediction(unknown_user: str, features: set[str],
    bestie_dict: dict[str, set[str]],
    feat_dict: dict[str, dict[str, str]]) -> dict[str, list[str]]:
    """Predicts a set of features for a user based on the features that their
    friends have.

    ### Parameters:
        unknown_user (str): The user who you are predicting attributes for.
        features (set): A set of features to predict.
        bestie_dict (dict): A dictionary of sets of friends of each individual
        in the social network.
        feat_dict (dict): A dictionary containing the known attributes for each
        user in the social network.

    ### Returns:
        predictions (dict): A dictionary of features with a predicted list of
        values for each feature.

    ### Examples:
    >>> friendly_prediction('glenn', {'favourite author', 'university'},
        {'kim': {'sandy', 'alex', 'glenn'}, 'sandy': {'kim', 'alex'},
        'alex': {'kim', 'sandy'}, 'glenn': {'kim'}},
        {'glenn': {'university': ''}, 'kim': {'favourite author': 'AA Milne'},
        'sandy': {'favourite author': 'JRR Tolkien',
        "university": "University of Melbourne"},
        'alex': {'favourite author': 'AA Milne',
        'university': 'Monash University'}})
    {'university': ['Monash University', 'University of Melbourne'],
    'favourite author': ['AA Milne']}

    >>> friendly_prediction('kim', {'university'},
        {'kim': {'sandy', 'alex', 'glenn'}, 'sandy': {'kim', 'alex'},
        'alex': {'kim', 'sandy'}, 'glenn': {'kim'}},
        {'glenn': {'university': ''}, 'kim': {'favourite author': 'AA Milne'},
        'sandy': {'favourite author': 'JRR Tolkien',
        "university": "University of Melbourne"},
        'alex': {'favourite author': 'AA Milne',
        'university': 'Monash University'}})
    {'university': ['', 'Monash University', 'University of Melbourne']}

    >>> friendly_prediction('kim', {'birthplace'},
        {'kim': {'sandy', 'alex', 'glenn'}, 'sandy': {'kim', 'alex'},
        'alex': {'kim', 'sandy'}, 'glenn': {'kim'}},
        {'glenn': {'university': ''}, 'kim': {'favourite author': 'AA Milne'},
        'sandy': {'favourite author': 'JRR Tolkien',
        "university": "University of Melbourne"},
        'alex': {'favourite author': 'AA Milne',
        'university': 'Monash University'}})
    {'birthplace': []}
    """

    predictions = {}

    for feature in features:
        # Start by searching through besties for the features
        prediction = predict_feature(feature, friend_besties(unknown_user, bestie_dict), feat_dict)

        # If there are still features that weren't able to be predicted, search
        # through second-besties instead
        if prediction is None:
            prediction = predict_feature(feature, friend_second_besties(unknown_user, bestie_dict),
                                         feat_dict)
        predictions[feature] = prediction if prediction is not None else []

    return predictions
